"""Natural parameter stepping strategy.

This module provides the natural parameter stepping strategy for the continuation package.
"""

from .base import _NaturalParameterStep

__all__ = [
    "_NaturalParameterStep",
]